@@ -24,8 +24,8 @@
  * @author Sam Meder
  *
  * $RCSfile: globus_i_callout.h,v $
- * $Revision: 1.13.14.1 $
- * $Date: 2012/02/13 14:31:05 $
+ * $Revision: 1.13.14.2 $
+ * $Date: 2013/06/06 05:42:45 $
  */
 
 #include "globus_callout.h"
@@ -228,6 +228,9 @@ typedef struct globus_i_callout_data_s
     char *                              file;
     char *                              symbol;
     struct globus_i_callout_data_s *    next;
+    char **                             env_args;
+    int                                 num_env_args;
+    globus_bool_t                       mandatory;
 } globus_i_callout_data_t;
 
 #endif
