@@ -18,9 +18,9 @@ static
 globus_version_t local_version = 
 {
     2,
-    2,
-    1329143466,
+    4,
+    1370497367,
     83
 };
 
-static const char flavor[] = "gcc64dbg";
+static const char flavor[] = "gcc64";
